@@ -244,7 +244,8 @@ gst_gl_filter_fixate_caps (GstBaseTransform * bt,
   GstStructure *ins, *outs;
   const GValue *from_par, *to_par;
   GValue fpar = { 0, }, tpar = {
-  0,};
+    0,
+  };
 
   othercaps = gst_caps_make_writable (othercaps);
   othercaps = gst_caps_truncate (othercaps);
@@ -458,7 +459,7 @@ gst_gl_filter_fixate_caps (GstBaseTransform * bt,
         goto done;
       }
 
-      /* If all this failed, keep the height that was nearest to the orignal
+      /* If all this failed, keep the height that was nearest to the original
        * height and the nearest possible width. This changes the DAR but
        * there's not much else to do here.
        */
@@ -611,7 +612,7 @@ gst_gl_filter_caps_remove_size (GstCaps * caps)
 }
 
 static GstCaps *
-gst_gl_filter_set_caps_features (const GstCaps * caps,
+gst_gl_filter_ensure_caps_contains_features (const GstCaps * caps,
     const gchar * feature_name)
 {
   GstCaps *ret = gst_caps_copy (caps);
@@ -619,8 +620,12 @@ gst_gl_filter_set_caps_features (const GstCaps * caps,
   guint i = 0;
 
   for (i = 0; i < n; i++) {
-    gst_caps_set_features (ret, i,
-        gst_caps_features_from_string (GST_CAPS_FEATURE_MEMORY_GL_MEMORY));
+    GstCapsFeatures *f = gst_caps_get_features (ret, i);
+    if (!gst_caps_features_is_any (f)) {
+      if (!gst_caps_features_contains (f, feature_name)) {
+        gst_caps_features_add (f, GST_CAPS_FEATURE_MEMORY_GL_MEMORY);
+      }
+    }
   }
 
   return ret;
@@ -651,7 +656,7 @@ gst_gl_filter_transform_caps (GstBaseTransform * bt,
         direction, caps, NULL);
 
     result =
-        gst_gl_filter_set_caps_features (tmp,
+        gst_gl_filter_ensure_caps_contains_features (tmp,
         GST_CAPS_FEATURE_MEMORY_GL_MEMORY);
     gst_caps_unref (tmp);
     tmp = result;
@@ -731,6 +736,7 @@ gst_gl_filter_set_caps (GstBaseTransform * bt, GstCaps * incaps,
 {
   GstGLFilter *filter;
   GstGLFilterClass *filter_class;
+  GstGLTextureTarget from_target, to_target;
 
   filter = GST_GL_FILTER (bt);
   filter_class = GST_GL_FILTER_GET_CLASS (filter);
@@ -740,12 +746,38 @@ gst_gl_filter_set_caps (GstBaseTransform * bt, GstCaps * incaps,
   if (!gst_video_info_from_caps (&filter->out_info, outcaps))
     goto wrong_caps;
 
+  {
+    GstStructure *in_s = gst_caps_get_structure (incaps, 0);
+    GstStructure *out_s = gst_caps_get_structure (outcaps, 0);
+
+    if (gst_structure_has_field_typed (in_s, "texture-target", G_TYPE_STRING))
+      from_target =
+          gst_gl_texture_target_from_string (gst_structure_get_string (in_s,
+              "texture-target"));
+    else
+      from_target = GST_GL_TEXTURE_TARGET_2D;
+
+    if (gst_structure_has_field_typed (out_s, "texture-target", G_TYPE_STRING))
+      to_target =
+          gst_gl_texture_target_from_string (gst_structure_get_string (out_s,
+              "texture-target"));
+    else
+      to_target = GST_GL_TEXTURE_TARGET_2D;
+
+    if (to_target == GST_GL_TEXTURE_TARGET_NONE
+        || from_target == GST_GL_TEXTURE_TARGET_NONE)
+      /* invalid caps */
+      goto wrong_caps;
+  }
+
   if (filter_class->set_caps) {
     if (!filter_class->set_caps (filter, incaps, outcaps))
       goto error;
   }
 
   gst_caps_replace (&filter->out_caps, outcaps);
+  filter->in_texture_target = from_target;
+  filter->out_texture_target = to_target;
 
   GST_DEBUG_OBJECT (filter, "set_caps %dx%d in %" GST_PTR_FORMAT
       " out %" GST_PTR_FORMAT,
@@ -758,7 +790,7 @@ gst_gl_filter_set_caps (GstBaseTransform * bt, GstCaps * incaps,
 /* ERRORS */
 wrong_caps:
   {
-    GST_WARNING ("Wrong caps");
+    GST_WARNING ("Wrong caps - could not understand input or output caps");
     return FALSE;
   }
 error:
@@ -772,7 +804,12 @@ gst_gl_filter_propose_allocation (GstBaseTransform * trans,
     GstQuery * decide_query, GstQuery * query)
 {
   GstGLFilter *filter = GST_GL_FILTER (trans);
-  GstGLContext *context = GST_GL_BASE_FILTER (filter)->context;
+  GstGLContext *context =
+      gst_gl_base_filter_get_gl_context (GST_GL_BASE_FILTER (filter));
+  if (context == NULL) {
+    return FALSE;
+  }
+
   GstCaps *caps;
   GstVideoInfo info;
   guint size;
@@ -800,34 +837,38 @@ gst_gl_filter_propose_allocation (GstBaseTransform * trans,
     gst_buffer_pool_config_set_params (config, caps, size, 0, 0);
 
     if (!gst_buffer_pool_set_config (pool, config)) {
-      g_object_unref (pool);
+      gst_object_unref (pool);
       goto config_failed;
     }
   }
 
   gst_query_add_allocation_pool (query, pool, size, 1, 0);
   if (pool)
-    g_object_unref (pool);
+    gst_object_unref (pool);
 
   if (context->gl_vtable->FenceSync)
     gst_query_add_allocation_meta (query, GST_GL_SYNC_META_API_TYPE, 0);
 
+  gst_object_unref (context);
   return TRUE;
 
   /* ERRORS */
 no_caps:
   {
     GST_DEBUG_OBJECT (trans, "no caps specified");
+    gst_object_unref (context);
     return FALSE;
   }
 invalid_caps:
   {
     GST_DEBUG_OBJECT (trans, "invalid caps specified");
+    gst_object_unref (context);
     return FALSE;
   }
 config_failed:
   {
     GST_DEBUG_OBJECT (trans, "failed setting config");
+    gst_object_unref (context);
     return FALSE;
   }
 }
@@ -835,7 +876,6 @@ config_failed:
 static gboolean
 gst_gl_filter_decide_allocation (GstBaseTransform * trans, GstQuery * query)
 {
-  GstGLContext *context;
   GstBufferPool *pool = NULL;
   GstStructure *config;
   GstCaps *caps;
@@ -851,7 +891,11 @@ gst_gl_filter_decide_allocation (GstBaseTransform * trans, GstQuery * query)
           query))
     return FALSE;
 
-  context = GST_GL_BASE_FILTER (trans)->context;
+  GstGLContext *context =
+      gst_gl_base_filter_get_gl_context (GST_GL_BASE_FILTER (trans));
+  if (context == NULL) {
+    return FALSE;
+  }
 
   if (gst_query_get_n_allocation_pools (query) > 0) {
     gst_query_parse_nth_allocation_pool (query, 0, &pool, &size, &min, &max);
@@ -889,6 +933,7 @@ gst_gl_filter_decide_allocation (GstBaseTransform * trans, GstQuery * query)
     gst_query_add_allocation_pool (query, pool, size, min, max);
 
   gst_object_unref (pool);
+  gst_object_unref (context);
 
   return TRUE;
 }
@@ -896,8 +941,8 @@ gst_gl_filter_decide_allocation (GstBaseTransform * trans, GstQuery * query)
 /**
  * gst_gl_filter_filter_texture:
  * @filter: a #GstGLFilter
- * @inbuf: an input buffer
- * @outbuf: an output buffer
+ * @input: an input buffer
+ * @output: an output buffer
  *
  * Calls filter_texture vfunc with correctly mapped #GstGLMemorys
  *
@@ -906,8 +951,8 @@ gst_gl_filter_decide_allocation (GstBaseTransform * trans, GstQuery * query)
  * Since: 1.4
  */
 gboolean
-gst_gl_filter_filter_texture (GstGLFilter * filter, GstBuffer * inbuf,
-    GstBuffer * outbuf)
+gst_gl_filter_filter_texture (GstGLFilter * filter, GstBuffer * input,
+    GstBuffer * output)
 {
   GstGLFilterClass *filter_class;
   GstMemory *in_tex, *out_tex;
@@ -916,7 +961,7 @@ gst_gl_filter_filter_texture (GstGLFilter * filter, GstBuffer * inbuf,
 
   filter_class = GST_GL_FILTER_GET_CLASS (filter);
 
-  if (!gst_video_frame_map (&gl_frame, &filter->in_info, inbuf,
+  if (!gst_video_frame_map (&gl_frame, &filter->in_info, input,
           GST_MAP_READ | GST_MAP_GL)) {
     ret = FALSE;
     goto inbuf_error;
@@ -929,7 +974,7 @@ gst_gl_filter_filter_texture (GstGLFilter * filter, GstBuffer * inbuf,
     goto unmap_out_error;
   }
 
-  if (!gst_video_frame_map (&out_frame, &filter->out_info, outbuf,
+  if (!gst_video_frame_map (&out_frame, &filter->out_info, output,
           GST_MAP_WRITE | GST_MAP_GL)) {
     ret = FALSE;
     goto unmap_out_error;
@@ -1001,6 +1046,15 @@ gst_gl_filter_transform (GstBaseTransform * bt, GstBuffer * inbuf,
   if (out_sync_meta)
     gst_gl_sync_meta_set_sync_point (out_sync_meta, context);
 
+  /* since gl api is async operation, when return from transform()
+   * function, basetransform will unref input buffer immidiately,
+   * but gpu may still reading input buffer for rendering. Add parent
+   * buffer meta to hold one reference of inbuf, this can avoid this
+   * buffer sync problem.
+   */
+  if (inbuf != outbuf)
+    gst_buffer_add_parent_buffer_meta (outbuf, inbuf);
+
   return ret ? GST_FLOW_OK : GST_FLOW_ERROR;
 }
 
@@ -1076,6 +1130,7 @@ _draw_with_shader_cb (GstGLFilter * filter, GstGLMemory * in_tex,
 {
   GstGLContext *context = GST_GL_BASE_FILTER (filter)->context;
   GstGLFuncs *gl = context->gl_vtable;
+  guint gl_target;
 
 #if GST_GL_HAVE_OPENGL
   if (gst_gl_context_get_gl_api (context) & GST_GL_API_OPENGL) {
@@ -1086,9 +1141,10 @@ _draw_with_shader_cb (GstGLFilter * filter, GstGLMemory * in_tex,
 
   _get_attributes (filter);
   gst_gl_shader_use (filter->default_shader);
+  gl_target = gst_gl_texture_target_to_gl (filter->in_texture_target);
 
   gl->ActiveTexture (GL_TEXTURE1);
-  gl->BindTexture (GL_TEXTURE_2D, gst_gl_memory_get_texture_id (in_tex));
+  gl->BindTexture (gl_target, gst_gl_memory_get_texture_id (in_tex));
 
   gst_gl_shader_set_uniform_1i (filter->default_shader, "tex", 1);
   gst_gl_shader_set_uniform_1f (filter->default_shader, "width",
@@ -1218,6 +1274,7 @@ gst_gl_filter_draw_fullscreen_quad (GstGLFilter * filter)
 
     if (gl->GenVertexArrays)
       gl->BindVertexArray (0);
-    _unbind_buffer (filter);
+    else
+      _unbind_buffer (filter);
   }
 }
